import logging

from typing import List

from ktdk.asserts.matchers import IsNotEmpty
from ktdk.tasks.build_task import BuildTask

log = logging.getLogger(__name__)


class RawCompilerTask(BuildTask):
    BASE_PARAMS = ['executable', 'files', 'compiler', 'use_default_options']

    def __init__(self, executable: str = 'bin_out', files: List = None,
                 compiler=None, use_default_options=True, **kwargs):
        super().__init__(**kwargs)
        self.files = files
        self._compiler = compiler
        self.executable = executable
        self._use_default_options = use_default_options
        self.add_tags('compile')

    @property
    def compiler(self):
        if self._compiler is None:
            self._compiler = self.context.config.get('cpp_compiler')
        return self._compiler

    def compile_run(self, ):
        found_files = self._file_utils.glob_files(self.files, source='workspace')
        self.asserts.require(found_files, matcher=IsNotEmpty("No input files to compile!"))
        str_files = [str(file_to_str) for file_to_str in found_files]
        build_dir = self.get_build_dir()
        exec_path = build_dir / self.executable
        if self.compiler_options and self._use_default_options:
            log.debug(f">>>>> [OPT] Compiler options: {self.compiler_options}")
            self.args.extend(self.compiler_options)
        self.args += ['-o', str(exec_path)]
        self.args += str_files
        log.info(f"[COMPILE] Compile [{self.executable}] with args {self.args}")
        self.execute()
        self.process_result(exec_path)

    def _run(self, *args, **kwargs):
        self._cmd = self.compiler
        log.info(f'[BUILD] Compilation for {self.namespace} - {self._cmd}')
        old_dir = self.move_to_dir(self.target)
        self.compile_run()
        self.move_to_dir(old_dir)

    def process_result(self, exec_path):
        self.context.config.add_test('exec', {self.executable: exec_path})

    def get_build_dir(self):
        return self.context.paths.workspace_path('build', create=True)

    @property
    def compiler_options(self):
        return []


class CCompilerTask(RawCompilerTask):
    TOOL_NAME = 'c_raw'

    @property
    def compiler_options(self):
        return self.context.config['c_compiler_flags']

    @property
    def compiler(self):
        if self._compiler is None:
            self._compiler = self.context.config.get('c_compiler')
        return self._compiler


class CPPCompilerTask(RawCompilerTask):
    TOOL_NAME = 'cpp_raw'

    @property
    def compiler(self):
        if self._compiler is None:
            self._compiler = self.context.config.get('cpp_compiler')
        return self._compiler

    @property
    def compiler_options(self):
        return self.context.config['cpp_compiler_flags']



import junitparser

from ktdk.asserts.checks.general import CheckTask


class AbstractXUnitSuiteCheck(CheckTask):
    @property
    def suite(self) -> junitparser.TestSuite:
        return self.context.config['xunit_suite']


class AbstractXUnitCaseCheck(CheckTask):
    @property
    def case(self) -> junitparser.TestCase:
        return self.context.config['xunit_case']


class XUnitSuiteFailsCheck(AbstractXUnitSuiteCheck):
    CHECK_NAME = 'xunit_fails'

    @property
    def _asserted_object(self):
        return self.suite.failures


class XUnitSuiteErrorsCheck(AbstractXUnitSuiteCheck):
    CHECK_NAME = 'xunit_errors'

    @property
    def _asserted_object(self):
        return self.suite.errors


class XUnitSuiteNokCheck(AbstractXUnitSuiteCheck):
    CHECK_NAME = 'xunit_nok'

    @property
    def _asserted_object(self):
        return self.suite.errors + self.suite.failures


class XUnitCaseCheck(AbstractXUnitCaseCheck):
    CHECK_NAME = 'xunit_case'

    @property
    def _asserted_object(self):
        return self.case.result



import pytest
import yaml

from ktdk.declarative import resolvers
from ktdk.tasks.fs import CopyFiles

FS_SCHEMA = """
tool: copy
test_files:
  - '*.c'
  - '../*.h'
submission:
  - pattern: '*.c'
    output_subdir: 'src'
"""


@pytest.fixture
def resolver():
    def _resolve(text):
        return resolvers.FSTasksResolver(yaml.safe_load(text))

    return _resolve


def test_simple_fs_resolver(resolver):
    entity = resolver(FS_SCHEMA).resolve()
    assert isinstance(entity, list)
    assert len(entity) == 3

    test_files0 = entity[0]
    assert isinstance(test_files0, CopyFiles)
    assert test_files0.pattern == ['*.c']

    test_files1 = entity[1]
    assert isinstance(test_files1, CopyFiles)
    assert test_files1.pattern == ['../*.h']

    submission = entity[2]
    assert isinstance(submission, CopyFiles)
    assert submission.pattern == ['*.c']
    assert submission.output_subdir == 'src'



import logging
from pathlib import Path

from typing import List

log = logging.getLogger(__name__)


class FileUtils:
    def __init__(self, context):
        self.context = context

    def _build_dir(self, source='workspace', subdir='', path=None) -> Path:
        if path:
            return Path(path)
        result = self.context.paths.get_dir(source)
        subdir = subdir if isinstance(subdir, list) else [subdir]
        return result / list_to_path(subdir)

    def glob_files(self, pattern, source='workspace', subdir='', path=None) -> List[Path]:
        location_dir: Path = self._build_dir(source=source, subdir=subdir, path=path)
        result = []
        if isinstance(pattern, str):
            pattern = [pattern]
        for item in pattern:
            result += location_dir.glob(item)
        log.debug(f"[GLOB] For pattern {pattern} in {location_dir} found: {result}")
        return result

    def list_directory_content(self, source='workspace', subdir='', path=None) -> List['str']:
        location_dir: Path = self._build_dir(source=source, subdir=subdir, path=path)
        import os
        return [dI for dI in os.listdir(str(location_dir))]

    def list_type_in_directories(self, source='workspace', subdir='', path=None, predicate=None):
        location_dir: Path = self._build_dir(source=source, subdir=subdir, path=path)
        content = self.list_directory_content(source=source, subdir=subdir, path=path)
        if predicate is None:
            return content
        return [c for c in content if predicate(location_dir.joinpath(c))]

    def exist_files(self, pattern, where='workspace', subdir='') -> bool:
        return len(self.glob_files(pattern=pattern, source=where, subdir=subdir)) >= 1

    def process_files(self, pattern, source='submission', target='workspace', action=None,
                      overwrite=True, subdir='', subdir_source='', subdir_target='',
                      transform=None, relative=True, **kwargs):
        subdir_target = Path(subdir or '') / (subdir_target or '')
        subdir_source = Path(subdir or '') / (subdir_source or '')
        location_dir: Path = self._build_dir(source=source, subdir=subdir)
        found = self.glob_files(pattern, source=source, subdir=subdir_source)
        paths = transform_path(found, transform)
        for found_file in paths:
            base_dir = location_dir if relative else None
            output_path = self.output_file(file_path=found_file, base_dir=base_dir,
                                           target=target, subdir_target=subdir_target)
            log.debug(f"[PROC] Manipulation ({action.__name__}): "
                      f"{found_file} -> {output_path}")
            create_dirs(output_path.parent)
            if not overwrite and output_path.exists():
                log.debug(f"[SKIP] File already exist: {output_path}")
            else:
                action(found_file, output_path, **kwargs)

    def output_file(self, file_path, target='workspace', subdir_target='', base_dir=None):
        file_path = Path(file_path)
        sub_path = file_path.name
        if base_dir:
            sub_path = file_path.relative_to(base_dir)
        output_dir = self._build_dir(source=target, subdir=subdir_target)
        return output_dir / sub_path


def transform_path(paths, action=None):
    action = action or (lambda x: x)
    return [action(p) for p in paths]


def create_dirs(path: Path, parent=False):
    if parent:
        path = path.parent
    if not path.exists():
        log.debug(f"[DIR] Dir not exists, creating: {path.parent}")
        path.mkdir(parents=True)


def list_to_path(parts: List['str']) -> Path:
    path = Path()
    for pth in parts:
        path /= pth
    return path



import os
import shutil
from pathlib import Path

import pytest

from ktdk import KTDK, Test
from ktdk.asserts import checks, matchers
from ktdk.tasks.cpp import CMakeBuildTask, CatchCheckCasesAndComputePoint, ValgrindCommand, catch
from ktdk.tasks.fs import CopyFiles, ExistFiles
from tests.paths import TEST_RESOURCES_BASE


def create_cmake(src, *names):
    content = """
    cmake_minimum_required (VERSION 3.0)
    project (kontr_tests)
    
    add_compile_options(-std=c++11)
    
    """
    src = Path(src)
    for name in names:
        content += f"add_executable({name} {name}.cpp)\n"
    cmake_file = src.joinpath('CMakeLists.txt')
    cmake_file.write_text(content)


@pytest.fixture
def workspace_dir(tmpdir):
    workspace = tmpdir.mkdir('workspace')
    return workspace


@pytest.fixture
def test_files_dir(tmpdir):
    test_files = tmpdir.mkdir('test_files')
    return test_files


@pytest.fixture
def submission_dir(tmpdir):
    submission = tmpdir.mkdir('submission')
    return submission


@pytest.fixture()
def prepared_sources(submission_dir):
    path = os.path.join(submission_dir, 'src')
    shutil.copytree(str(TEST_RESOURCES_BASE / 'catch'), path)
    create_cmake(path, 'tests')
    return submission_dir


@pytest.fixture
def cmake():
    cmake = CMakeBuildTask(source='src', cwd='src')
    cmake.require_that(checks.TaskResultCheck(matcher=matchers.ResultPassed()))
    return cmake


@pytest.fixture()
def file_tasks():
    result = [
        ExistFiles(pattern='tests.cpp', subdir='src', source='submission', required=True),
        CopyFiles(pattern='*.*', subdir='src', source='submission'),
        CopyFiles('CMakeLists.txt', source='submission'),
    ]

    return result


@pytest.fixture()
def catch_test():
    catch_test = Test(name="mini_test", desc="Super catch test")
    catch_task = catch.CatchRunTestsOneByOneTask(executable='tests', executor=ValgrindCommand)
    catch_test.add_task(catch_task)
    catch_test.check_that(CatchCheckCasesAndComputePoint(max_points=3), after_tasks=True)
    return catch_test


@pytest.fixture()
def naostro(catch_test):
    naostro = Test(name="naostro", desc="Test nanecisto", tags=['naostro'])
    naostro.add_test(catch_test)
    return naostro


@pytest.fixture()
def ktdk(workspace_dir, test_files_dir, submission_dir, cmake, naostro, file_tasks):
    ktdk = KTDK(
        test_files=test_files_dir,
        submission=submission_dir,
        workspace=workspace_dir,
        devel=True
    )
    ktdk.suite.add_task(*file_tasks)
    ktdk.suite.add_task(cmake)
    ktdk.suite.add_test(naostro)
    return ktdk


@pytest.mark.slow
@pytest.mark.scenario
@pytest.mark.catch
def test_catch_run(prepared_sources, ktdk):
    ktdk.invoke()
    assert ktdk.suite.result.effective.passed
    assert round(ktdk.suite.result.effective_points, 3) == 3



from typing import Optional

from ktdk.asserts.checks.general import CheckTask


class StyleResultMatchesCheck(CheckTask):
    CHECK_NAME = 'style_result'

    @property
    def _asserted_object(self) -> Optional[str]:
        if not self.context:
            return None
        return self.context.config['style_result']


class ClangTidyResultMatchesCheck(CheckTask):
    CHECK_NAME = 'tidy_result'

    @property
    def _asserted_object(self) -> Optional[str]:
        if not self.context:
            return None
        return self.context.config['tidy_result']



from typing import Dict

from ktdk import utils
from ktdk.core.tasks import Task


class AbortTask(Task):
    def __init__(self, message: str = None, **kwargs):
        super().__init__(**kwargs)
        self.message = message
        self.description = self.message or message
        self.name = self.name or 'abort_task'

    def _run(self, *args, **kwargs):
        self.asserts.abort(self.message)


class CheckTask(Task):
    CHECK_NAME = None

    @classmethod
    def check_register(cls) -> Dict:
        return utils.make_subclasses_register(CheckTask, 'CHECK_NAME')

    def __init__(self, matcher=None, **kwargs):
        super().__init__(**kwargs, checked=True)
        self.matcher = matcher

    @property
    def _asserted_object(self):
        return None

    def _run(self, *args, **kwargs):
        self.asserts.check(self._asserted_object, self.matcher)


class TaskResultCheck(CheckTask):
    CHECK_NAME = 'task_result'

    @property
    def _asserted_object(self):
        return self.parent.result.current


class TaskEffectiveResultCheck(CheckTask):
    CHECK_NAME = 'task_result_effective'

    @property
    def _asserted_object(self):
        return self.parent.result.effective


class TestEffectiveResultCheck(CheckTask):
    CHECK_NAME = 'test_result_effective'

    @property
    def _asserted_object(self):
        return self.test.result.effective


class TestResultCheck(CheckTask):
    CHECK_NAME = 'test_result'
    @property
    def _asserted_object(self):
        return self.test.result.current


class AbstractExecResultMatchesTask(CheckTask):
    @property
    def exec_result(self):
        if not self.context:
            return None
        return self.context.config['exec_result']



import pytest
import yaml

from ktdk import KTDK, declarative

HELLO_DEF = """
config:
    timeout: 20
tags:
    - naostro
    - nanecisto

suite:
    name: hello
    description: Hello world intro program
    tasks:
        - fs:
            tool: exist
            submission:
                - 'hello.c'
            test_files:
                - 'solution.c'
            required: true
            
        - fs:
              tool: copy
              submission:
                  - 'hello.c'
              test_files:
                  - pattern: 'solution.c'
        - build:
            tool: c_raw
            executable: student_hello
            files:
                - hello.c
            required: true
        - build:
            tool: c_raw
            executable: teacher_hello
            files:
                - solution.c
            required: true
    tests:
        - name: student_hello
          description: Student's hello should write hello
          tags:
            - naostro
          tasks:
            - execute:
                executable: student_hello
                executor: valgrind
                checks:
                    - type: stdout
                      regex: 'Hello World!'
                    - type: stderr
                      equals: ''
                    - type: return_code
                      equals: 0
                      
        - name: teacher_hello
          description: Teacher's hello should write hello
          tags:
           - nanecisto
          tasks:
            - execute:
                executable: teacher_hello
                executor: valgrind
                checks:
                    - type: stdout
                      regex: 'Hello World!'
                    - type: stderr
                      equals: ''
                    - type: return_code
                      equals: 0
"""


def create_file(where, name, content=None):
    default = (
        """#include <stdio.h>

int main() {
  printf("Hello World!\\n");
  return 0;
}""")
    content = content or default
    path = where.join(f'{name}.c')
    path.write(content)
    return path


@pytest.fixture
def workspace_dir(tmpdir):
    workspace = tmpdir.mkdir('workspace')
    return workspace


@pytest.fixture
def test_files_dir(tmpdir):
    test_files = tmpdir.mkdir('test_files')
    return test_files


@pytest.fixture
def submission_dir(tmpdir):
    submission = tmpdir.mkdir('submission')
    return submission


@pytest.fixture()
def prepared_sources(submission_dir, test_files_dir):
    create_file(submission_dir, 'hello')
    create_file(test_files_dir, 'solution')
    return submission_dir


@pytest.fixture()
def ktdk(workspace_dir, test_files_dir, submission_dir) -> KTDK:
    ktdk = KTDK(
        test_files=test_files_dir,
        submission=submission_dir,
        workspace=workspace_dir
    )

    return declarative.load_suite(yaml.safe_load(HELLO_DEF), ktdk)


def test_decl_hello_test(ktdk, prepared_sources):
    ktdk.invoke()
    assert ktdk.suite.result.effective.passed



import logging
import os
from pathlib import Path
from typing import Type

from ktdk import utils
from ktdk.core.mixins import ToolTaskMixin
from ktdk.tasks.command_task import CommandTask

log = logging.getLogger(__name__)


class BuildTask(CommandTask, ToolTaskMixin):
    BASE_PARAMS = ['source_dir', 'target_dir', 'cwd']

    @classmethod
    def _base_class(cls):
        return BuildTask

    def __init__(self, source_dir=None, target_dir=None, cwd=None, **kwargs):
        super().__init__(**kwargs)
        self._source_dir = source_dir or ''
        self._target_dir = target_dir or 'build'
        self._cwd = cwd
        self.add_tags('build')

    @property
    def source(self):
        return self.base / self._source_dir

    @property
    def base(self):
        return self.context.paths.workspace

    @property
    def target(self):
        return self.base / self._target_dir

    def source_dir(self, source):
        self._source_dir = source
        return self

    def target_dir(self, target):
        self._target_dir = target
        return self

    def __create_dir(self, directory=None, create=True):
        directory = directory or self.target
        if create:
            utils.create_dirs(directory)

    def move_to_dir(self, directory: Path, create=True):
        directory = Path(directory)
        self.__create_dir(directory, create=create)
        old_dir = Path(os.getcwd())
        log.debug(f"[MOVE] Moving to dir: {directory} from {old_dir}")
        os.chdir(directory)
        return old_dir




import difflib
import io
import logging
from pathlib import Path
from typing import AnyStr, Dict, Optional, Union

import intelhex
import yaml

from .file_utils import FileUtils, create_dirs, list_to_path
from .flatters import flatten_all_tasks, flatten_tasks, flatten_tests
from .meta_utils import bind_class_var, load_class, make_subclasses_register, subclasses_by_attr
from .naming import normalize_name, unique_name

log = logging.getLogger(__name__)


def parse_yaml_config(file_path: Path) -> Optional[Dict]:
    """Loads the YAML file from the defined path
    Args:
        file_path: File path from which the YAML file should be loaded

    Returns(Dictionary): loaded yaml

    """
    file_path = Path(file_path)
    if not file_path.exists():
        log.warning(f"[YAML] Cannot load file - not exists: {file_path}")
        return {}
    with file_path.open('r') as stream:
        try:
            return yaml.safe_load(stream)
        except yaml.YAMLError as exc:
            # TODO Throw an exception
            log.error(f"[PARSE] Error for {file_path}: {exc}")
    return {}


def save_yaml(file_path, config):
    """Saves the config to the config file
    Args:
        file_path: YAML file path
        config: Configuration holder

    """
    file_path = Path(file_path)
    with file_path.open('w') as stream:
        try:
            yaml.safe_dump(config, stream, default_flow_style=False)
        except yaml.YAMLError as exc:
            # TODO Throw an exception
            log.error(f"[SAVE] Error: {exc}")


def universal_reader(content: Union[Path, io.RawIOBase, str], raw=True) -> str:
    content = extract_content(content)
    if isinstance(content, Path):
        with content.open('rb') as fd:
            content = fd.read()
    if hasattr(content, 'read'):
        content = content.read()
    if raw:
        return content
    else:
        return content.decode('utf-8')


def dig_class(obj, *selector):
    current = obj
    for sel in selector:
        current = getattr(current, sel)
        if not current:
            return None
    return current


def try_decode(provided: AnyStr) -> Optional[str]:
    log.info(f"[DECODE] Decoding[{type(provided)}]: {provided}")
    if not hasattr(provided, 'decode'):
        return provided
    try:
        return provided.decode('utf-8')
    except UnicodeError:
        return None


def try_encode(provided: AnyStr) -> Optional[str]:
    if not hasattr(provided, 'encode'):
        return provided
    try:
        return provided.encode('utf-8')
    except UnicodeError:
        return None


def hex_diff(provided, expected):
    exp_hex = intelhex.IntelHex()
    prov_hex = intelhex.IntelHex()

    exp_hex.frombytes(expected)
    prov_hex.frombytes(provided)
    return intelhex.diff_dumps(exp_hex, prov_hex, name1='expected', name2='provided')


def get_context_diff(expected: AnyStr, provided: AnyStr, keep_ends=True):
    def _split_string(orig):
        if isinstance(orig, str):
            orig = orig.splitlines(keepends=keep_ends)
        return orig

    decoded_prov = try_decode(provided)
    decoded_exp = try_decode(expected)
    if decoded_exp is not None and decoded_prov is not None:
        expected = _split_string(decoded_exp)
        provided = _split_string(decoded_prov)
        return difflib.context_diff(expected, provided, fromfile='expected', tofile='provided')
    return hex_diff(provided, expected)


def extract_content(content: Union[str, dict, bytes]):
    """Extracted content
    Args:
        content: Content input
    Returns:

    """
    if isinstance(content, dict):
        return Path(content.get('file'))
    if isinstance(content, str):
        return io.StringIO(content)
    if isinstance(content, bytes):
        return io.BytesIO(content)
    if isinstance(content, Path):
        return content
    if isinstance(content, io.RawIOBase):
        return content
    return content




import importlib
import inspect
from typing import List, Type, Dict


def bind_class_var(cls, name: str) -> List[str]:
    cache_name = f'_CACHE_{name}'
    if not hasattr(cls, cache_name):
        parents = [klass for klass in inspect.getmro(cls) if hasattr(klass, name)]
        collection = []
        for klass in parents:
            collection += getattr(klass, name)
        setattr(cls, cache_name, collection)
    return getattr(cls, cache_name)


def subclasses_by_attr(cls: Type, attr: str) -> List:
    subs = cls.__subclasses__()
    result = []
    if hasattr(cls, attr):
        result.append(cls)
    for klass in subs:
        result.extend(subclasses_by_attr(klass, attr))
    return result


def make_subclasses_register(cls: Type, attr: str) -> Dict:
    subs = subclasses_by_attr(cls, attr)
    register = {}
    for klass in subs:
        selector = getattr(klass, attr)
        if selector is None:
            continue
        if not isinstance(selector, List):
            selector = [selector]
        for select in selector:
            register[select] = klass
    return register


def load_class(klass):
    parts = klass.split('.')
    class_name = parts[-1]
    module_path = ".".join(parts[:-1])
    module = importlib.import_module(module_path)
    my_class = getattr(module, class_name)
    return my_class



import logging
from typing import Dict, TYPE_CHECKING

from ktdk import utils
from ktdk.core import errors
from ktdk.runtime.context import Context
from ktdk.utils import FileUtils

if TYPE_CHECKING:
    from ktdk.core.tasks import Task

log = logging.getLogger(__name__)


class ContextMixin:
    @property
    def context(self) -> Context:
        ctx = getattr(self, '_context', None)
        if ctx is None:
            raise errors.ContextIsNotInitializedError(self.__class__.__name__)
        return ctx

    @property
    def has_context(self) -> bool:
        return hasattr(self, '_context')

    def inject_context(self, context):
        setattr(self, '_context', context)


class CheckedMixin:
    def check_that(self, task: 'Task', points_multiplier: float = 0.0, after_tasks: bool = False):
        """Adds the check task
        Args:
            after_tasks:
            points_multiplier: If this task fails, how much from the test points should be discarded
            task(Task): Task instance
        """
        from .tasks import TaskType
        if getattr(task, '_type') == TaskType.NORMAL:
            setattr(task, '_type', TaskType.CHECKED)
        setattr(task, '_points_multiplier', points_multiplier)
        if hasattr(self, 'add_after') and after_tasks:
            self.add_after(task)
        else:
            self.add_task(task)
        return self

    def require_that(self, task: 'Task', points_multiplier: float = 0.0, after_tasks: bool = False):
        """Requires a subtask to pass
        Args:
            after_tasks: It should be executed after the tasks execution
            points_multiplier: If this task fails, how much from the test points should be discarded
            task(Task):
        """
        from .tasks import TaskType
        self.check_that(task, points_multiplier=points_multiplier, after_tasks=after_tasks)
        task.type = TaskType.REQUIRED
        return self

    def abort(self, message: str, points_multiplier: float = 0):
        from ktdk.asserts.checks.general import AbortTask
        log.warning(f"[ABORT] Aborted {self.name}: {message}")
        self.require_that(AbortTask(message=message), points_multiplier=points_multiplier)


class FileUtilsMixin:
    @property
    def _file_utils(self) -> FileUtils:
        if not hasattr(self, '__file_utils') and hasattr(self, 'context'):
            setattr(self, '__file_utils', FileUtils(self.context))
        return getattr(self, '__file_utils')


class ToolTaskMixin:
    TOOL_NAME = None

    @classmethod
    def _base_class(cls):
        return ToolTaskMixin

    @classmethod
    def tools_register(cls) -> Dict:
        return utils.make_subclasses_register(cls._base_class(), 'TOOL_NAME')


class ExecutorMixin:
    EXECUTOR = None

    @classmethod
    def exec_register(cls) -> Dict:
        return utils.make_subclasses_register(ExecutorMixin, 'EXECUTOR')



import abc
import collections
import json
import logging
import os
from pathlib import Path
from typing import Union

from ktdk import utils

CONFIG_PATH = Path(__file__).parent / 'resources' / 'config'

log = logging.getLogger(__name__)


class _CustomJSONEncode(json.JSONEncoder):
    def default(self, o):
        return str(o)


def _dump_json(obj) -> str:
    return json.dumps(obj, cls=_CustomJSONEncode, indent=4)


def _load_env_var(key: str, val: str):
    if not key.startswith('KTDK_'):
        return None
    name = key[5:].lower()
    if val in ['True', 'true']:
        val = True
    if val in ['False', 'false']:
        val = False
    return {name: val}


def _load_env_vars() -> dict:
    params = {}
    for (key, val) in os.environ.items():
        parsed = _load_env_var(key, val)
        if parsed:
            params.update(parsed)
    return params


def _get_subdir(where: Path, sub_path, create=False):
    full_path = where / Path(sub_path)
    if create:
        utils.create_dirs(full_path)
    return full_path


class ConfigPaths:
    """Context dirs wrapper for paths
    """

    def __init__(self, config: 'ConfigPropMixin'):
        """Creates instance of the paths dirs
        Args:
            config(ConfigPropMixin): Context instance
        """
        self._container = config

    @property
    def workspace(self) -> Path:
        """Gets workspace path
        Returns(Path): Workspace path
        """
        return Path(self._container['workspace'])

    @property
    def test_files(self) -> Path:
        """Test files path
        Returns(Path): Test file path
        """
        return Path(self._container['test_files'])

    @property
    def submission(self) -> Path:
        """Submission files path
        Returns:

        """
        return Path(self._container['submission'])

    def submission_path(self, sub_path: Union[str, Path], create=False) -> Path:
        return _get_subdir(self.submission, sub_path, create)

    def test_path(self, sub_path: Union[str, Path], create=False) -> Path:
        return _get_subdir(self.test_files, sub_path, create)

    def result_path(self, sub_path: Union[str, Path], create=False) -> Path:
        return _get_subdir(self.results, sub_path, create)

    def workspace_path(self, sub_path: Union[str, Path], create=False) -> Path:
        return _get_subdir(self.workspace, sub_path, create)

    @property
    def results(self) -> Path:
        """Gets results path
        Returns(Path): Results path

        """
        results_dir = self._container['results']
        if not results_dir:
            results_dir = self.workspace / 'results'
        return Path(results_dir)

    def get_dir(self, name) -> Path:
        return getattr(self, name)

    def resolve(self, name) -> Path:
        return getattr(self, name, name)

    def save_result(self, file_path: Union[str, Path], content, raw=True) -> Path:
        path = self.results / Path(file_path)
        utils.create_dirs(path=path.parent)
        log.debug(f"[SAVE] Save content to the results file: {file_path}")
        if raw:
            path.write_bytes(content)
        else:
            path.write_text(content, encoding='utf-8')
        return path

    def save_test_result(self, test, file_name: str, content: str, raw=True):
        path = utils.list_to_path(test.namespace_parts) if test else Path()
        path /= file_name
        return self.save_result(path, content=content, raw=raw)

    def save_task_result(self, task, file_name: str, content, raw=True):
        fn = task.task_namespace + "-" + file_name
        if len(fn) > 255:
            fn = file_name
        return self.save_test_result(task.test, fn, content=content, raw=raw)


class ConfigPropMixin:
    @property
    def paths(self) -> ConfigPaths:
        if not hasattr(self, '__paths'):
            setattr(self, '__paths', ConfigPaths(self))
        return getattr(self, '__paths')

    @property
    def devel(self) -> bool:
        return self.__props__.get('devel', False)

    @property
    def kill(self) -> bool:
        return self.__props__.get('kill', False)

    @property
    def clean(self) -> bool:
        return self.__props__.get('clean', False)

    @property
    def submission_config(self) -> dict:
        return self.__props__['submission_config']

    @submission_config.setter
    def submission_config(self, value):
        self.__props__['submission_config'] = value

    @property
    @abc.abstractmethod
    def __props__(self) -> dict:
        return {}


class Config(collections.MutableMapping, ConfigPropMixin):
    def __len__(self) -> int:
        return len(self.container)

    def __iter__(self):
        return iter(self.container)

    def __delitem__(self, item):
        del self.container[item]

    def __init__(self, **params):
        self._container = params or {}

    def load(self):
        self.load_envs()  # highest priority
        save_cfg = dict(self._container)
        self._load_default()
        self._load_provided()
        self.update(dictionary=save_cfg, override=True)

    def log_config(self):
        log.info(f"[CFG] Config: {_dump_json(self.container)}")

    def update(self, dictionary, override=False):
        if override:
            self.container.update(dictionary)
        else:
            self._container = {**dictionary, **self._container}

    @property
    def container(self) -> dict:
        return self._container

    def __getitem__(self, item):
        return self._container.get(item)

    def __setitem__(self, key, value):
        self._container[key] = value

    def _load_default(self):
        """Loads the default config from the CONFIG
        """
        self.load_yaml(CONFIG_PATH / 'defaults.yml', override=False)

    def load_yaml(self, file_path, override=False):
        """Loads the YAML config from any location
        Args:
            override: Whether to override provided config
            file_path: Yaml location file path
        """
        from ktdk import utils
        config = utils.parse_yaml_config(file_path)
        log.debug(f"[LOAD] CFG {file_path}: {config}")
        self.update(config, override=override)

    def load_envs(self):
        """Loads the excluded_params from the env variables
        """
        envs = _load_env_vars()
        self.update(envs, override=True)

    def load_dict(self, params: dict):
        """Loads the dictionary
        Args:
            params:
        Returns:
        """
        self._container.update(params)

    def save_yaml(self, file_path):
        """Saves the configuration to the YAML file
        Args:
            file_path: YAML file path
        """
        from ktdk import utils
        utils.save_yaml(file_path, self)

    @property
    def __props__(self):
        return self.container

    def _load_provided(self):
        if not self.paths.test_files:
            return

        provided_path = self.paths.test_files / self.get('ktdk_config_path', '.ktdk_config.yml')
        if provided_path.exists():
            log.info(f"[CFG] Loading provided KTDK config for project: {provided_path}")
            self.load_yaml(provided_path, override=True)
        else:
            log.debug(f"[CFG] KTDK config not provided: {provided_path}]")



import logging
from typing import Dict, List

from ktdk import Task, Test, utils
from ktdk.core import errors
from ktdk.declarative import registry
from ktdk.tasks.raw.executable import ExecutableTask

log = logging.getLogger(__name__)


class BaseResolver:
    SELECTOR = None
    BASE_PARAMS = []
    KLASS = None

    @classmethod
    def base_params(cls):
        return utils.bind_class_var(cls, 'BASE_PARAMS')

    @classmethod
    def registered(cls):
        return utils.make_subclasses_register(BaseResolver, 'SELECTOR')

    def __init__(self, definition: Dict):
        self._definition = definition

    @property
    def definition(self):
        return self._definition

    def _resolve(self, klass=None):
        return self.definition

    def _after_resolution(self, entity):
        pass

    def resolve(self):
        klass = self._get_class(self.__class__.KLASS)
        entity = self._resolve(klass)
        self._after_resolution(entity)
        log.debug(f"[RESOLV] Resolved using {self.__class__.__name__}: {entity}")
        return entity

    def _get_class(self, klass):
        if isinstance(self.definition, dict) and 'class' in self.definition.keys():
            klass = utils.load_class(self.definition['class'])
        return klass


class ConfigResolver(BaseResolver):
    SELECTOR = 'config'


class TagsResolver(BaseResolver):
    SELECTOR = 'tags'


class TestBaseResolver(BaseResolver):
    SELECTOR = 'test'
    KLASS = Test

    @property
    def _test_params(self) -> Dict:
        excluded = ['before', 'after', 'before_all', 'after_all', 'tasks', 'tests']
        return {key: self.definition[key] for key in self.definition.keys() if key not in excluded}

    def _resolve(self, klass=None):
        return klass(**self._test_params)


class SuiteResolver(TestBaseResolver):
    SELECTOR = 'suite'


class TaskBaseResolver(BaseResolver):
    SELECTOR = 'task'
    KLASS = Task

    @property
    def definition(self) -> Dict:
        if len(self._definition.keys()) == 1:
            name = self.__class__.SELECTOR
            return self._definition[name]
        return self._definition

    def _task_params(self, klass) -> Dict:
        params = klass.base_params()
        return {key: self.definition[key] for key in self.definition.keys() if key in params}

    def _resolve(self, klass=None):
        return klass(**self._task_params(klass))

    def _after_resolution(self, entity):
        self._resolve_checks(entity)

    def _resolve_checks(self, task: Task):
        checks = self.definition.get('checks')
        if checks:
            check_list = [_resolve_check(check) for check in checks]
            for check in check_list:
                task.check_that(check)


class FSTasksResolver(TaskBaseResolver):
    SELECTOR = 'fs'

    def _resolve_tool(self):
        tool_name = self.definition.get('tool', 'copy')
        tool = registry.fs_tool_get(tool_name)
        if not tool:
            raise errors.DeclRegistryNotFoundError('file tools', tool_name)
        return tool

    def _resolve(self, klass=None):
        tool = self._resolve_tool()
        result = []
        for f_type in ['test_files', 'submission']:
            file_tasks = self._resolve_files(f_type, tool)
            result.extend(file_tasks)
        return result

    def _resolve_files(self, files_type, tool) -> List[Task]:
        result = []
        files_list = self.definition[files_type]
        required = self.definition.get('required') or False
        checked = self.definition.get('checked') or False
        for i, file_item in enumerate(files_list):
            params = {'source': files_type, 'required': required, 'checked': checked}
            if isinstance(file_item, str):
                params['pattern'] = file_item
            else:
                params.update(file_item)
            tool_name = tool.TOOL_NAME
            if isinstance(tool_name, list):
                tool_name = next(iter(tool_name))
            params['name'] = params.get('name', f'{tool_name}_{files_type}_{i}')

            instance = tool(**params)
            result.append(instance)
        return result


class BuildTaskResolver(TaskBaseResolver):
    SELECTOR = 'build'

    def _resolve_tool(self):
        tool_name = self.definition.get('tool')
        tool = registry.build_get(tool_name)
        if not tool:
            raise errors.DeclRegistryNotFoundError('build', tool_name)
        return tool

    def _resolve(self, klass=None) -> Task:
        tool = self._resolve_tool()
        instance = tool(**self._task_params(tool))
        return instance


class ExecuteTaskResolver(TaskBaseResolver):
    SELECTOR = 'execute'

    def _resolve_executor(self):
        tool_name = self.definition.get('executor')
        tool = registry.executor_get(tool_name)
        return tool

    def _resolve(self, klass=None) -> Task:
        executor = self._resolve_executor()
        exec_tool = self._get_class(ExecutableTask)
        params = self._task_params(exec_tool)
        params['executor'] = executor
        instance = exec_tool(**params)
        return instance


def _resolve_matcher(check):
    for (key, val) in check.items():
        item = registry.get_matcher(key)
        if item is not None:
            return item(val) if val is not None else item()
    raise errors.DeclMatcherNotExistError(check)


def _resolve_check(check: Dict):
    check_type = check.get('type')
    if not check_type:
        raise errors.DeclAttributeMissingError('\'type\' for check')

    check_class = registry.checks_get(check_type)

    if check_class is None:
        raise errors.DeclRegistryNotFoundError('checks', check_type)
    matcher = _resolve_matcher(check)

    return check_class(matcher=matcher)



import logging

from ktdk import utils
from ktdk.core.tests import Test
from ktdk.utils.flatters import flatten_all_tasks, flatten_tests

log = logging.getLogger(__name__)


def _get_tests_with_result(tests, result, res_type='effective'):
    return [t for t in tests if utils.dig_class(t.result, res_type, result)]


def _count_for_tasks(test):
    tasks = flatten_all_tasks(test)
    return _count_for_collection(tasks)


def _count_col_for_type(res_type, coll, with_checked=False):
    count = {}
    all_len = len(coll)
    filtered = _get_tests_with_result(coll, res_type)
    len_filter = len(filtered)
    if len_filter == 0:
        return None
    count['all'] = len_filter
    count['ratio'] = len_filter / all_len
    current_tests = _get_tests_with_result(coll, res_type='current', result=res_type)
    count['current'] = len(current_tests)
    count['current_ratio'] = len(current_tests) / all_len

    if with_checked:
        count_checked = len([test for test in coll if test.has_checked])
        checked = [test for test in filtered if test.has_checked]
        count['checked'] = len(checked)
        count['checked_ratio'] = len(checked) / count_checked

    return count


def _count_for_collection(coll, with_checked=False):
    all_len = len(coll)
    count = dict(all=all_len)
    if with_checked:
        count_checked = len([test for test in coll if test.has_checked])
        count['all_checked'] = count_checked
    for res_type in ['passed', 'failed', 'skipped', 'errored']:
        for_type = _count_col_for_type(res_type=res_type, coll=coll, with_checked=with_checked)
        if for_type:
            count[res_type] = for_type
    return count


def _get_tests_names(tests):
    result = {}

    def __names(res_type):
        sel = _get_tests_with_result(tests, res_type)
        if sel:
            result[res_type] = [selected.namespace for selected in sel]

    for rtype in ['failed', 'errored', 'skipped']:
        __names(res_type=rtype)
    return result


def get_task_info(task):
    return dict(
        name=task.name,
        namespace=task.namespace,
        tags=list(task.tags),
        description=task.description,
        result=dict(effective=task.result.effective.state, current=task.result.current.state)
    )


def get_test_info(test, with_tasks=False):
    test_info = dict(name=test.name, namespace=test.namespace, description=test.description,
                     tags=list(test.tags),
                     effective_tags=list(test.effective_tags), points=test.points,
                     result=dict(effective=test.result.effective.state,
                                 current=test.result.current.state,
                                 reduced_points=test.result.reduced_points,
                                 effective_points=test.result.effective_points))
    if test.result.current.nok or with_tasks:
        extract_tasks_info(test_info, test)
    return test_info


def extract_tasks_info(test_info, test):
    tasks = flatten_all_tasks(test)
    test_info['tasks'] = [get_task_info(task) for task in tasks]
    nok = {task.task_namespace: task.result.current.state for task in tasks if
           task.result.current.nok}
    test_info['tasks_nok'] = nok


def stat_test(test) -> dict:
    """Statistics of the tests
    Args:
        test(Test): Root test

    Returns(dict): computed statistics

    """
    all_tests = flatten_tests(test)
    points = test.result.effective_points
    result = dict(final_points=round(points, 4),
                  result=test.result.effective.state)
    context = test.context or None
    with_tasks = context.config.get('dump_passed') if context else False
    result['tests_count'] = _count_for_collection(all_tests, with_checked=True)
    result['tasks_count'] = _count_for_tasks(test)
    tests = _get_tests_with_result(all_tests, result='passed')
    result['passed_tests'] = [test.namespace for test in tests]
    tests = _get_tests_with_result(all_tests, result='failed')
    result['failed_tests'] = [test.namespace for test in tests]
    result['all_tests'] = [get_test_info(test, with_tasks=with_tasks) for test in all_tests]
    return result



from pathlib import Path

PROJECT_BASE = Path(__file__).parent.parent
TEST_BASE = Path(__file__).parent
TEST_RESOURCES_BASE = TEST_BASE.joinpath('resources')
TEST_RES_DECL = TEST_RESOURCES_BASE / 'declarative'



import logging
import typing
from copy import deepcopy
from pathlib import Path
from typing import Iterator

from ktdk import utils

from ktdk.config import ConfigPaths, ConfigPropMixin
from ktdk.runtime.tags import TagsEvaluator
from ktdk.utils.basic import BasicObject

log = logging.getLogger(__name__)


def deep_merge(source, destination):
    """Deep merge
    Examples:
    >>> a = { 'first' : { 'all_rows': { 'pass': 'dog', 'number': '1' } } }
    >>> b = { 'first' : { 'all_rows': { 'fail': 'cat', 'number': '5' } } }
    >>> deep_merge(b, a) == {'first':{'all_rows':{'pass': 'dog', 'fail': 'cat', 'number': '5'}}}
    True
    """
    for key, value in source.items():
        if isinstance(value, dict):
            # get node or create one
            node = destination.setdefault(key, {})
            deep_merge(value, node)
        else:
            destination[key] = value

    return destination


def _get_subdir(where: Path, sub_path, create=False):
    full_path = where / Path(sub_path)
    if create:
        utils.create_dirs(full_path)
    return full_path


class ContextConfig(ConfigPropMixin, typing.MutableMapping):
    @property
    def __props__(self) -> dict:
        return self.all

    def __init__(self, suite_config=None, test_config=None, task_config=None):
        self._suite = suite_config if suite_config is not None else {}
        self._test = test_config if test_config is not None else {}
        self._task = task_config if task_config is not None else {}

    @property
    def suite(self):
        return self._suite

    @property
    def test(self):
        return self._test

    @property
    def task(self):
        return self._task

    @property
    def submission(self) -> dict:
        return self.get('submission_config')

    @property
    def all(self):
        cfg = {}
        cfg = deep_merge(self.suite, cfg)
        cfg = deep_merge(self.test, cfg)
        cfg = deep_merge(self.task, cfg)
        return cfg

    def __set_any(self, which, name, value):
        collection = getattr(self, which)
        if name in collection:
            log.debug(f"[CTX] Overriding {which}['{name}']: ({collection[name]}) by ({value})!")
        else:
            log.trace(f"[CTX] SET: {which}['{name}'] = {value}")
        collection[name] = value

    def __add_any(self, which, name, value):
        collection = getattr(self, which)
        old_value = collection.get(name, None)
        new_value = deepcopy(value)

        if isinstance(value, list):
            new_value = old_value + value if old_value else deepcopy(value)

        if isinstance(value, dict):
            new_value = deep_merge(old_value or {}, new_value)

        log.trace(f"[CTX] Adding ({which}['{name}']): {old_value} + {value} -> {new_value}")
        collection[name] = new_value

    def add_suite(self, name, value):
        self.__add_any('suite', name, value)

    def add_test(self, name, value):
        self.__add_any('test', name, value)

    def add_task(self, name, value):
        self.__add_any('task', name, value)

    def set_suite(self, name, value):
        self.__set_any('suite', name, value)

    def set_test(self, name, value):
        self.__set_any('test', name, value)

    def set_task(self, name, value):
        self.__set_any('task', name, value)

    def clone(self, clone_test=False):
        test_config = deepcopy(self.test) if clone_test else self.test
        task_config = deepcopy(self.task)
        config = ContextConfig(suite_config=self.suite,
                               test_config=test_config,
                               task_config=task_config)
        return config

    def __getitem__(self, item):
        return self.all.get(item)

    def get(self, name):
        return self.all.get(name)

    def __str__(self):
        return str(self.all)

    def __repr__(self):
        return str(self.all)

    def __delitem__(self, v) -> None:
        # Todo implement for any collection
        del self.all[v]

    def __len__(self) -> int:
        return len(self.all)

    def __iter__(self) -> Iterator:
        return self.all.__iter__()

    def __setitem__(self, k, v) -> None:
        self.all[k] = v


class Context(BasicObject, typing.MutableMapping):
    def __init__(self, suite_config=None, test_config=None, task_config=None, config=None):
        super().__init__()
        self._config = config or ContextConfig(suite_config=suite_config,
                                               test_config=test_config,
                                               task_config=task_config)
        self._tags_evaluator = TagsEvaluator(self.config['tags'], self.config['registered_tags'])
        self._paths = ConfigPaths(self.config)

    @property
    def tags(self):
        return self._tags_evaluator

    @property
    def devel(self):
        return self.config.all.get('devel', False)

    @property
    def paths(self) -> ConfigPaths:
        return self._paths

    @property
    def config(self):
        return self._config

    @config.setter
    def config(self, value):
        self._config = value

    def clone(self, clone_test=False):
        return Context(config=self.config.clone(clone_test=clone_test))

    def to_dict(self):
        return self.config

    def __getitem__(self, item):
        """Gets an item form the context
        """
        return self.config[item]

    def __setitem__(self, k, v) -> None:
        self.config[k] = v

    def __delitem__(self, v) -> None:
        del self.config[v]

    def __len__(self) -> int:
        return len(self.config)

    def __iter__(self) -> Iterator:
        return self.config.__iter__()



class KTDKError(Exception):
    def __init__(self, msg: str = None, **kwargs):
        super().__init__()
        self.msg = msg

    def __str__(self):
        return f"{self.__class__.__name__}: {self.msg}"


class KTDKAssertionError(KTDKError):
    pass


class RequireFailedError(KTDKAssertionError):
    pass


class KillCheckError(KTDKAssertionError):
    pass


class RequiredTaskFailed(KTDKAssertionError):
    pass


class TaskRunFailed(KTDKError):
    pass


class ContextIsNotInitializedError(KTDKError):
    def __init__(self, class_name: str):
        message = f'Context has not been initialized yet for class: {class_name}'
        super(ContextIsNotInitializedError, self).__init__(msg=message)


class DeclarativeError(KTDKError):
    def __init__(self, msg: str = None):
        super(DeclarativeError, self).__init__(msg=msg)


class DeclSchemaError(KTDKError):
    def __init__(self, msg: str = None):
        super(DeclSchemaError, self).__init__(msg=msg)


class DeclAttributeMissingError(DeclSchemaError):
    def __init__(self, attribute: str = None):
        msg = f"Attribute missing in the schema: {attribute}"
        super(DeclSchemaError, self).__init__(msg=msg)


class DeclMatcherNotExistError(DeclarativeError):
    def __init__(self, schema=None):
        msg = f"Matcher not found for the check: {schema}"
        super(DeclMatcherNotExistError, self).__init__(msg=msg)


class DeclRegistryNotFoundError(DeclarativeError):
    def __init__(self, which=None, name=None):
        msg = f"{which} not found: {name}"
        super(DeclRegistryNotFoundError, self).__init__(msg=msg)


class FileNotExists(KTDKError):
    pass



from pathlib import Path

from ktdk.asserts.matchers import GeneralMatcher


class FileExists(GeneralMatcher):
    MATCHER_NAME = ['exists', 'exist']

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs, symbol='is')
        self.message = self.message or 'File not Exists'

    def predicate(self, orig):
        orig = Path(orig)
        return orig.exists()



import logging

from ktdk.asserts import matchers
from ktdk.core.mixins import FileUtilsMixin
from ktdk.core.tasks import Task
from ktdk.tasks.command import Command, CommandResult

log = logging.getLogger(__name__)


class CommandTask(Task, FileUtilsMixin):
    BASE_PARAMS = ['command', 'args', 'executor', 'command_config', 'output_name', 'cwd',
                   'input']

    def __init__(self, command=None, args=None, executor=Command, input=None,
                 command_config=None, output_name=None, cwd=None, **kwargs):
        super().__init__(**kwargs)
        self.add_tags('command')
        self.name = self.name or command
        self.executor = executor or Command
        self._cmd = command
        self._cmd_config = command_config or {}
        self.args = args or []
        self._output_name = output_name
        if cwd is not None:
            self._cmd_config['cwd'] = cwd
        if input is not None:
            self._cmd_config['input'] = input

    def build_args(self, *args):
        return [*args, *self.args]

    @property
    def command_path(self) -> str:
        """Name of the command, if not provided, it will try to use the <task_name>_command
        form the config
        Returns(str): name or path to the command
        """
        return self._cmd or self.context.config.get(self.name + '_command')

    @property
    def command_config(self) -> dict:
        if self.has_context:
            self._cmd_config['timeout'] = self._cmd_config.get('timeout') or \
                                          self.__extract_timeout()
        return self._cmd_config

    def __extract_timeout(self):
        return self.context.config.get(self.name + '_timeout') or \
               self.context.config.get('timeout')

    def command(self):
        """Creates the command instance based on the executor
            Executor can be the Command or ValgrindCommand
        Returns:

        """
        path = str(self.command_path)
        args = self.build_args()
        return self.executor(path, args=args, output_name=self._output_name, **self.command_config)

    def execute(self, **kwargs) -> CommandResult:
        return self.command().set_task(self).invoke(**kwargs)

    def _run(self, *args, **kwargs):
        result = self.execute()
        if result.nok:
            log.warning(f"[CMD] Command failed {self.namespace}: {result}")
            self.add_tags('failed')
        else:
            self.add_tags('passed')
        self.asserts.check(result, matchers.CommandOK())
        self.context.config.set_test('command_result', result)



from typing import Type

from ktdk import tasks
from ktdk.asserts.checks import *
from ktdk.asserts.matchers import *
from ktdk.core.mixins import ExecutorMixin
from ktdk.tasks.fs.tasks import *


def full() -> Dict[str, Dict[str, Type]]:
    return {
        'build': build(),
        'checks': checks(),
        'file_tools': fs_tools(),
        'executors': executors(),
        'matchers': matchers(),
        'resolvers': resolvers()
    }


def build() -> Dict:
    return tasks.build_task.BuildTask.tools_register()


def checks() -> Dict:
    return CheckTask.check_register()


def fs_tools() -> Dict:
    return tasks.fs.AbstractFilesTask.tools_register()


def executors() -> Dict:
    return ExecutorMixin.exec_register()


def matchers() -> Dict:
    return GeneralMatcher.matchers_register()


def resolvers() -> Dict:
    from . import resolvers
    return resolvers.BaseResolver.registered()


def fs_tool_get(name: str):
    return fs_tools().get(name)


def checks_get(name: str):
    return checks().get(name)


def build_get(name: str):
    return build().get(name)


def executor_get(name: str):
    return executors().get(name)


def resolver_get(name):
    return resolvers().get(name)



import logging
from pathlib import Path
from typing import Collection, List, Set

from ktdk import Task, Test
from ktdk.asserts import checks, matchers
from ktdk.core.mixins import FileUtilsMixin
from ktdk.scenarios import FullScenario, Scenario
from ktdk.scenarios.simple import InOutFileScenario
from ktdk.tasks import cpp, fs
from ktdk.tasks.cpp import ValgrindCommand

log = logging.getLogger(__name__)


class CMiniSingleTaskScenario(FullScenario):
    """Execute just single CMiniTask scenario
    """

    def __init__(self, task_name: str, points=0.1, wd_subdir=None, executor=ValgrindCommand,
                 **kwargs):
        super().__init__(**kwargs)
        self.task_name = task_name
        self.points = points
        self.executor = executor
        self.wd_subdir = wd_subdir

    def checkout_teacher_files(self):
        log.info(f"[CHECKOUT] All teacher files for task {self.task_name}")
        copy = fs.CopyFiles(f"{self.task_name}/**/*.*", source='test_files', relative=True)
        self.add_task(copy)

    def checkout_student_files(self):
        source_file = f"{self.task_name}/source.c"
        self.require_that(fs.ExistFiles(source_file))
        copy = fs.CopyFiles(source_file, source='submission', relative=True)
        self.add_task(copy)

    def file_tasks(self):
        self.checkout_teacher_files()
        self.checkout_student_files()  # Needs to be in this order

    def compile_tasks(self):
        c_compile = cpp.CCompilerTask(executable=self.task_name,
                                      files=f'{self.task_name}/source.c',
                                      use_default_options=True)
        c_compile.require_that(checks.TaskResultCheck(matcher=matchers.ResultPassed()))
        c_compile.require_that(checks.ExecutableExists(executable=self.task_name))
        self.root_test.add_task(c_compile)
        return c_compile

    def run_tasks(self):
        params = dict(points=self.points)
        if self.wd_subdir:
            params['wd_subdir'] = self.wd_subdir

        subtests = CMiniSubtestForTask(self.task_name, **params)
        self.root_test.add_task(subtests)


class CMiniSubtestForTask(Task, FileUtilsMixin):
    def __init__(self, task_name, points=0.1, wd_subdir=None, **kwargs):
        super().__init__(**kwargs)
        self.task_name = task_name
        self.points = points
        self.wd_subdir = wd_subdir

    @property
    def base(self) -> Path:
        return self.context.paths.workspace / self.task_name

    def __build_paths(self, ext: str) -> List['Path']:
        return list(self.base.glob(f'*.{ext}'))

    @property
    def out_paths(self) -> List['Path']:
        return self.__build_paths('out')

    @property
    def in_paths(self) -> List['Path']:
        return self.__build_paths('in')

    @property
    def args_paths(self) -> List['Path']:
        return self.__build_paths('args')

    @property
    def err_paths(self) -> List['Path']:
        return self.__build_paths('err')

    def _run(self, *args, **kwargs):
        names = dict(
            outs=_resolve_names(self.out_paths),
            ins=_resolve_names(self.in_paths),
            errs=_resolve_names(self.err_paths),
            args=_resolve_names(self.args_paths)
        )

        all_names = set().union(*names.values())
        log.info(f"[MINI] All subtests: {all_names}")
        all_tests = [_build_test(name, names) for name in all_names]

        num = len(all_tests)
        for test in all_tests:
            self._run_one(test, num=num)

    def _in_out_test(self, name, points=0.0, stdin=None, stdout=None,
                     args=None, status_code=0, stderr=None, executor=ValgrindCommand):
        """Create instance of the in/out test
        Args:
            name(str): Test name
            points(float): Number of points for the test
            stdin(Path): File with stdin
            stdout(Path): Stdout file
            args(List['str']): List of arguments
            status_code(int): Status code
            stderr(Path): Stdout path
            executor: Executor instance

        Returns:

        """
        test = Test(name=name, desc=f'Subtest name: {name}', points=points)
        in_out_scenario = InOutFileScenario(
            executable=self.task_name, args=args, stdin=stdin, executor=executor,
            stdout=stdout, stderr=stderr, status_code=status_code, wd_subdir=self.wd_subdir)
        test.add_task(in_out_scenario)
        return test

    def _run_one(self, test: dict, num=0):
        def get_path(val):
            return self.base / val if val else None

        params = {key: get_path(value) for key, value in test.items() if key != 'name'}

        points = self.points / num
        if params.get('args') is not None:
            args_path: Path = params['args']
            if args_path.exists():
                args = args_path.read_text('utf-8').split('\n')
                params['args'] = [arg for arg in args if arg]

        test = self._in_out_test(test['name'], points=points, **params)
        self.test.add_test(test)


class CMiniAllTasksScenario(Scenario):
    def __init__(self, points=1, wd_to_mini_dir=False, **kwargs):
        super().__init__(**kwargs)
        self.points = points
        self.wd_to_mini_dir = wd_to_mini_dir

    def list_teacher_directories(self) -> Set[str]:
        return set(self._file_utils.list_type_in_directories(source='test_files',
                                                             predicate=Path.is_dir))

    def list_of_student_impl(self) -> Set['str']:
        files = self._file_utils.glob_files(pattern='*/source.c', source='submission')
        sources = {p.parent.name for p in files}
        return sources

    def _run(self, *args, **kwargs):
        teacher: set = self.list_teacher_directories()
        student: set = self.list_of_student_impl()
        intersection = student.intersection(teacher)
        num_tasks = len(teacher)
        ratio = self.points / float(num_tasks) if num_tasks else 0
        for task_name in intersection:
            test_for_mini_task = Test(name=task_name, desc=f"Test for task: {task_name}")
            task_scenario = CMiniSingleTaskScenario(task_name=task_name, points=ratio,
                                                    wd_subdir=task_name)
            test_for_mini_task.add_task(task_scenario)
            self.root_test.add_test(test_for_mini_task)


def _resolve_names(collection: Collection['Path']) -> Set['str']:
    return {p.resolve().stem for p in collection}


def _intersect_names(first: Set['str'], *sets) -> Set['str']:
    result = first
    for name_set in sets:
        result = result.intersection(name_set)
    return result


def _build_test(name, names):
    test = dict(
        stdout=f"{name}.out" if name in names['outs'] else None,
        stdin=f"{name}.in" if name in names['ins'] else None,
        stderr=f'{name}.err' if name in names['errs'] else None,
        args=f'{name}.args' if name in names['args'] else None,
        name=name,
    )
    return test



import json
import os
from typing import IO, Any

import yaml


class YAMLIncludeLoader(yaml.SafeLoader):
    """YAML Loader with `!include` constructor."""

    def __init__(self, stream: IO) -> None:
        """Initialise Loader."""

        try:
            self._root = os.path.split(stream.name)[0]
        except AttributeError:
            self._root = os.path.curdir

        super().__init__(stream)


def construct_include(loader: YAMLIncludeLoader, node: yaml.Node) -> Any:
    """Include file referenced at node."""

    filename = os.path.abspath(os.path.join(loader._root, loader.construct_scalar(node)))
    extension = os.path.splitext(filename)[1].lstrip('.')

    with open(filename, 'r') as f:
        if extension in ('yaml', 'yml'):
            return yaml.load(f, YAMLIncludeLoader)
        elif extension in ('json',):
            return json.load(f)
        else:
            return ''.join(f.readlines())



import logging
from typing import Any, Dict, List, Optional, Union

import yaml

from ktdk import KTDK, Task, Test
from ktdk.core import errors
from ktdk.declarative import registry, resolvers
from ktdk.declarative.yaml_extend import YAMLIncludeLoader, construct_include

yaml.add_constructor('!include', construct_include, YAMLIncludeLoader)

log = logging.getLogger(__name__)


def extract_params(definition: Dict, excluded: List = None):
    excluded = excluded or []
    return {key: definition[key] for key in definition.keys() if key not in excluded}


class DeclarativeBaseLoader:
    KLASS = None
    EXCLUDED_PARAMS = []

    @classmethod
    def klass(cls, *args, **kwargs):
        return cls.KLASS(*args, **kwargs)

    def __init__(self, definition: Dict, throws=False):
        self.definition = definition
        self._throws = throws
        self._errors = []

    def _add_error(self, err):
        log.error(f"[DECL] {err}")
        self._errors.append(err)

    def _resolve_entity(self, param: str = None, resolver=None):
        if resolver is None:
            resolver = registry.resolver_get(param)
        if param is None:
            param = resolver.SELECTOR
        entity = self.definition.get(param)
        if param in self.definition:
            return resolver(entity).resolve()
        return None

    def load(self) -> Any:
        pass


class DeclarativeSuiteLoader(DeclarativeBaseLoader):
    def __init__(self, definition: Dict, ktdk: KTDK):
        super().__init__(definition)
        self._ktdk = ktdk

    @property
    def ktdk(self) -> KTDK:
        return self._ktdk

    def load(self):
        imports = self._resolve_entity(param='imports')
        self._resolve_imports(imports)
        config = self._resolve_entity(param='config')
        self._ktdk.config.update(config, override=True)
        tags = self._resolve_entity(param='tags')
        self._ktdk.register_tags(*tags)
        self._load_suite()
        return self.ktdk

    def _load_suite(self):
        suite_schema = self.definition.get('suite')
        if not suite_schema:
            return None
        suite = DeclarativeTestLoader(definition=suite_schema).load()
        if suite:
            self._ktdk.suite = suite
            log.debug(f"[LOAD] Suite: {suite}")
        return suite

    def _resolve_imports(self, imports):
        pass


def _collect_tasks(collection):
    result = []
    for item in collection:
        loaded = DeclarativeTaskLoader(item).load()
        if isinstance(loaded, Task):
            loaded = [loaded]
        result.extend(loaded)
    return result


class DeclarativeTestLoader(DeclarativeBaseLoader):
    KLASS = Test

    def _create_instance(self) -> Optional[Test]:
        resolver = resolvers.TestBaseResolver
        test_resolvers = resolver.registered()
        if '_type' in self.definition.keys():
            found = test_resolvers.get(self.definition['_type'])
            if found:
                resolver = found
        try:
            test = resolver(self.definition).resolve()
            log.info(f"[LOAD] Test {test}")
            return test
        except errors.DeclarativeError as ex:
            self._add_error(ex)
        return None

    def load(self) -> Optional[Test]:
        if 'test' in self.definition:
            self.definition = self.definition['test']
        test = self._create_instance()
        if test is None:
            return None
        self._add_tasks_if_present(test)
        test.add_test(*self._load_tests())
        return test

    def _add_tasks_if_present(self, test):
        test.add_before(*self._load_tasks('before'))
        test.add_before(*self._load_tasks('before_each'), scope='each')
        test.add_after(*self._load_tasks('after'))
        test.add_before(*self._load_tasks('after_each'), scope='each')
        test.add_task(*self._load_tasks('tasks'))

    def _load_tasks(self, task_type) -> Optional[List[Task]]:
        collection = self.definition.get(task_type)
        if not collection:
            return []
        return _collect_tasks(collection)

    def _load_tests(self) -> Optional[List[Test]]:
        collection = self.definition.get('tests')
        if not collection:
            return []
        return [DeclarativeTestLoader(item).load() for item in collection]


class DeclarativeTaskLoader(DeclarativeBaseLoader):
    KLASS = Task

    TYPES = {}

    def _create_instance(self) -> Optional[Task]:
        resolver = self._get_resolver()
        try:
            task = resolver(self.definition).resolve()
            log.info(f"[LOAD] Task: {task}")
            return task
        except errors.DeclarativeError as ex:
            self._add_error(ex)
        return None

    def _get_resolver(self):
        resolver = resolvers.TaskBaseResolver
        task_resolvers = resolver.registered()
        def_keys = self.definition.keys()
        if len(def_keys) == 1:
            name = next(iter(def_keys))
            found = task_resolvers.get(name)
            if found:
                return found
        if '_type' in self.definition.keys():
            found = task_resolvers.get(self.definition['_type'])
            if found:
                return found
        return resolver

    def load(self) -> Union[Task, List]:
        task = self._create_instance()
        if isinstance(task, Task):
            task.add_task(*self._load_tasks())
            return task
        log.debug("Adding additional task to same level since the result of the resolver is list")
        return [*task, *self._load_tasks()]

    def _load_tasks(self) -> Optional[List[Task]]:
        collection = self.definition.get('tasks')
        if not collection:
            return []
        return _collect_tasks(collection)



import io

import pytest

from ktdk import KTDK, Test
from ktdk.scenarios.simple import InOutFileScenario
from ktdk.tasks.cpp import RawCompilerTask
from tests.utils import get_test_context


@pytest.fixture
def context(prepared_dir):
    config = dict(workspace=prepared_dir, devel=True)
    return get_test_context(suite_config=config, test_config={})


@pytest.fixture()
def workspace(tmpdir):
    return tmpdir.mkdir('workspace')


@pytest.fixture()
def ktdk_instance(workspace):
    return KTDK(
        workspace=workspace,
        devel=True
    )


@pytest.fixture
def prepared_dir(tmpdir, workspace):
    c_file = workspace.join('main.c')
    c_file.write("""
        #include <stdio.h>\n
        #include <string.h>\n
        int main(int argc, char *argv[]) {
            for(int i = 1; i < argc; i++) {
                puts(argv[i]);
            }
            char buffer[256];
            fgets(buffer, 255, stdin);
            puts(buffer);
            return 0; 
        }
    """)
    return workspace


@pytest.fixture()
def scenario(prepared_dir):
    test = Test(name="Simple test")
    test.add_task(RawCompilerTask(compiler='gcc', executable='hello', files=['main.c']))
    scenario = InOutFileScenario(
        executable='hello',
        stdout=io.StringIO("hello\n"),
        stdin=io.StringIO("hello"),
    )
    test.add_task(scenario)
    return test


@pytest.mark.slow
@pytest.mark.scenario
@pytest.mark.compile
def test_compile_tasks(ktdk_instance, scenario):
    ktdk_instance.suite.add_test(scenario)
    result = ktdk_instance.run()
    assert result
    assert ktdk_instance.suite.result.effective.passed



import logging
from pathlib import Path

from ktdk import utils
from ktdk.tasks.command import Command

log = logging.getLogger(__name__)


class ValgrindCommand(Command):
    EXECUTOR = 'valgrind'

    def __init__(self, command, args=None, **kwargs):
        """Creates Valgrind command instance
        Args:
            command(str): Command name
            args(list): Command arguments
        """
        super().__init__(command, args=args, **kwargs)
        self._vargs = []

    @property
    def valgrind_log_file(self) -> str:
        """Gets name of the valgrind log file
        Returns(str): Valgrind log file
        """
        name = self.output_file_name(suffix='valgrind.log')
        test_path = utils.list_to_path(self.task.test.namespace_parts)
        path_name: Path = self.context.paths.results / test_path / name
        utils.create_dirs(path_name.parent)
        return str(path_name)

    @property
    def valgrind_command(self):
        """Valgrind command with arguments and sets log file
        Returns(list): valgrind command with valgrind arguments
        """
        args = []
        if self.context:
            valgrind_file = self.valgrind_log_file
            args.append(f'--log-file={valgrind_file}')
            args.extend(self.context.config.get('valgrind_args'))
        cmd = ['valgrind', *args, *self._vargs]
        return cmd

    def add_vargs(self, *args):
        """Adds the valgrind arguments
        Args:
            *args: Valgrind arguments
        Returns(ValgrindCommand): Valgrind command instance
        """
        self._vargs.extend(args)
        return self

    @property
    def executable_command(self):
        """Executable command with arguments
        Returns(list): Executable command with arguments list
        """
        return [self.command, *self.args]

    @property
    def full_command(self):
        """Full command list
        Returns(list): Full command with valgrind and executable
        """
        return [*self.valgrind_command, *self.executable_command]

    def process(self, result):
        """Processes the result
        Args:
            result: Command execution result
        Returns: Processes result
        """
        process_result = super().process(result)
        process_result.outputs['valgrind_log'] = self.valgrind_log_file

        self.context.config.set_test('valgrind_log', self.valgrind_log_file)
        return process_result



import pytest
import yaml

from ktdk.core.tasks import TaskType
from ktdk.declarative import resolvers
from ktdk.tasks import cpp

BUILD_SCHEMA = """
tool: c_raw
executable: hello
# Files are relative to workspace
files:
  - hello.c
required: true
"""


@pytest.fixture
def resolver():
    def _resolve(text):
        return resolvers.BuildTaskResolver(yaml.safe_load(text))

    return _resolve


def test_simple_build_resolver(resolver):
    entity = resolver(BUILD_SCHEMA).resolve()
    assert isinstance(entity, cpp.compiler.CCompilerTask)
    assert entity.executable == 'hello'
    assert entity.files == ['hello.c']
    assert entity.type == TaskType.REQUIRED



import os
import shutil

import pytest

from ktdk import KTDK, Test
from ktdk.scenarios.c import mini
from tests.paths import TEST_RESOURCES_BASE


@pytest.fixture
def workspace_dir(tmpdir):
    workspace = tmpdir.mkdir('workspace')
    return workspace


@pytest.fixture
def test_files_dir(tmpdir):
    test_files = tmpdir.mkdir('test_files')
    return test_files


@pytest.fixture
def submission_dir(tmpdir):
    submission = tmpdir.mkdir('submission')
    return submission


@pytest.fixture()
def prepared_sources(submission_dir):
    path = os.path.join(submission_dir, 'src')
    shutil.copytree(str(TEST_RESOURCES_BASE / 'c_mini'), path)
    return path


@pytest.fixture()
def mini_test():
    mini_test = Test(name="MiniTest", desc="Super mini test")
    mini_scenario = mini.CMiniSingleTaskScenario('args', points=0.2)
    mini_test.add_task(mini_scenario)
    return mini_test


@pytest.fixture()
def naostro(mini_test):
    naostro = Test(name="naostro", desc="Test naostro", tags=['naostro'])
    naostro.add_test(mini_test)
    return naostro


@pytest.fixture()
def ktdk(workspace_dir, test_files_dir, prepared_sources, naostro):
    ktdk = KTDK(
        test_files=prepared_sources,
        submission=prepared_sources,
        workspace=workspace_dir,
        devel=True
    )
    ktdk.suite.add_test(naostro)
    return ktdk


@pytest.mark.slow
@pytest.mark.scenario
@pytest.mark.catch
def test_catch_run(prepared_sources, ktdk):
    ktdk.invoke()
    assert ktdk.suite.result.effective.passed
    assert round(ktdk.suite.result.effective_points, 3) == 0.2



import logging
from typing import Union
from xml.etree import ElementTree

from junitparser import JUnitXml, TestCase, TestSuite

from ktdk.asserts.checks.general import TaskResultCheck
from ktdk.asserts.matchers import IsNone, ResultPassed
from ktdk.core.tasks import Task
from ktdk.core.tests import Test

log = logging.getLogger(__name__)


class JUnitParseTask(Task):
    def __init__(self, junit_file=None, **kwargs):
        super().__init__(**kwargs)
        self._junit_file = junit_file
        self._root = None

    @property
    def xml_root(self):
        if self._root is None:
            self._root = JUnitXml.fromfile(str(self._junit_file))
        return self._root

    def parse_xunit(self):
        for suite in self.xml_root:
            log.debug(f"[HANDLE] SUITE: {suite}")
            self.process_the_suite(suite)

    def process_the_suite(self, suite):
        log.debug(f"[JUNIT] Test Suite {suite.name}: {suite}")
        self.context.config.set_test('xunit_suite', suite)
        suite_test = create_test(suite, test=self.test)
        suite_test.add_tags('junit_suite')
        for case in suite:
            _handle_case(case, suite_test=suite_test)
        return suite_test

    def _run(self, *args, **kwargs):
        log.debug(f"[JUNIT] Parse: {self.test.namespace}")
        try:
            return self.parse_xunit()
        except ElementTree.ParseError as ex:
            self.asserts.abort(f"There is an error the JUnit parsing: {ex}")


class _ProcessTestCaseTask(Task):
    def __init__(self, case: TestCase, **kwargs):
        super().__init__(**kwargs)
        self.case = case

    def _run(self, *args, **kwargs):
        self.asserts.check(self.case.result, matcher=IsNone())
        self.context.config.add_test('xunit_case', self.case)


def _handle_case(test_case: TestCase, suite_test: Test):
    log.debug(f"[JUNIT] Test Case {test_case.name} [{test_case.result}]: {test_case}")
    case_test = build_test_case(test_case=test_case)
    suite_test.add_test(case_test)
    return suite_test


def create_test(test_case: Union[TestCase, TestSuite], test: Test = None):
    test = Test() if test is None else test
    test.description = test_case.name or 'Undefined name'
    test.name = test_case.name or 'Undefined name'
    log.debug(f"[JUNIT] Create test: {test_case}")
    return test


def build_test_case(test_case: TestCase):
    test = create_test(test_case=test_case)
    test.add_tags('junit_case')
    process_task = _ProcessTestCaseTask(test_case)
    process_task.check_that(TaskResultCheck(matcher=ResultPassed()))
    test.add_task(process_task)
    return test



from typing import Dict

import pytest
import yaml

from ktdk import Test, declarative

JUST_TEST_YAML = """
name: just_test
description: 'what ever'
tags: ['some', 'tags']
points: 5
"""

TEST_W_TESTS_YAML = """
name: just_test
description: 'what ever'
tags: ['some', 'tags']
points: 5
tests:
    - name: Foo
      points: 0.5
    - name: Bar
      points: 0.5
"""


@pytest.fixture()
def just_test_definition() -> Dict:
    return yaml.safe_load(JUST_TEST_YAML)


@pytest.fixture()
def test_w_test_definition() -> Dict:
    return yaml.safe_load(TEST_W_TESTS_YAML)


def test_just_test_params_serializer(just_test_definition):
    test_loader = declarative.DeclarativeTestLoader(just_test_definition)
    loaded: Test = test_loader.load()
    assert loaded.name == 'just_test'
    assert loaded.description == 'what ever'
    assert loaded.tags == {'some', 'tags'}
    assert loaded.points == 5


def test_test_with_tests_serializer(test_w_test_definition):
    test_loader = declarative.DeclarativeTestLoader(test_w_test_definition)
    loaded: Test = test_loader.load()
    assert loaded.name == 'just_test'
    assert loaded.description == 'what ever'
    assert loaded.tags == {'some', 'tags'}
    assert loaded.points == 5
    assert len(loaded.tests) == 2
    assert loaded.tests[0].name == 'foo'
    assert loaded.tests[1].name == 'bar'



import pytest
import yaml

from ktdk.declarative import resolvers
from ktdk.tasks.cpp import ValgrindCommand
from ktdk.tasks.raw.executable import ExecutableTask

EXEC_SCHEMA = """
executable: engine
executor: valgrind
input:
    file: simple_test.in
checks:
    - type: stdout
      equals: "Engine works!"
    - type: stderr
      match: 'some regex'
    - type: return_code
      greater: 10
"""


@pytest.fixture
def resolver():
    def _resolve(text):
        return resolvers.ExecuteTaskResolver(yaml.safe_load(text))

    return _resolve


def test_simple_build_resolver(resolver):
    entity = resolver(EXEC_SCHEMA).resolve()
    assert isinstance(entity, ExecutableTask)
    assert entity.executable_name == 'engine'
    assert entity.executor == ValgrindCommand
    assert entity.command_config['input'] == dict(file='simple_test.in')
    assert len(entity.tasks) == 3



import logging
import re

from ktdk.asserts import matchers
from ktdk.asserts.checks.general import AbstractExecResultMatchesTask

log = logging.getLogger(__name__)


class ExecutableExists(AbstractExecResultMatchesTask):
    CHECK_NAME = 'exec_exist'

    def __init__(self, executable: str, **kwargs):
        super().__init__(**kwargs)
        self._executable = executable

    def _run(self, *args, **kwargs):
        self.asserts.require(self.context, matcher=matchers.DictHasKey('exec'))
        executables = self.context['exec']
        self.asserts.require(executables, matcher=matchers.DictHasKey(self._executable))
        exec_path = executables[self._executable]
        self.asserts.require(exec_path, matcher=matchers.FileExists())


class ReturnCodeMatchesCheck(AbstractExecResultMatchesTask):
    CHECK_NAME = 'return_code'

    def _run(self, *args, **kwargs):
        self.asserts.check(self.exec_result.return_code, self.matcher)


class StdOutMatchesCheck(AbstractExecResultMatchesTask):
    CHECK_NAME = 'stdout'

    def _run(self, *args, **kwargs):
        self.asserts.check(self.exec_result.stdout.content, self.matcher)


class StdErrMatchesCheck(AbstractExecResultMatchesTask):
    CHECK_NAME = 'stderr'

    def _run(self, *args, **kwargs):
        self.asserts.check(self.exec_result.stderr.content, self.matcher)


class ValgrindPassedCheck(AbstractExecResultMatchesTask):
    CHECK_NAME = 'valgrind'

    REGEX = re.compile(r'ERROR SUMMARY: (\d+) errors')

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @property
    def points_multiplier(self):
        if self._points_multiplier:
            return self._points_multiplier

        return self.context.config['valgrind_penalization'] or 0

    @property
    def valgrind_log(self) -> str:
        return self.context.config['valgrind_log']

    def __read_log_content(self):
        valgrind_log = self.valgrind_log
        if valgrind_log is None:
            return None
        valgrind_log = self.context.paths.results / valgrind_log
        return valgrind_log.read_text(encoding='utf-8')

    def _run(self, *args, **kwargs):
        self.asserts.check(self._asserted_zero_errors(), self.matcher)

    def _asserted_zero_errors(self):
        content = self.__read_log_content()
        if content is None:
            log.warning(f"[VALGRIND] Valgrind content is not set for {self.namespace}")
            return False
        passed = True
        errors = None
        for line in content.splitlines():
            match = ValgrindPassedCheck.REGEX.search(line)
            if match:
                errors = match.group(1)
                passed = int(errors) == 0
        if not passed:
            log.warning(f"[VALGRIND] Log {self.namespace}:\n{content}")
            log.warning(f"[VALGRIND] Found errors in Valgrind: {errors}")
        else:
            log.debug(f'[VALGRIND] Log {self.namespace}:\n{content}')
        return passed



import logging
import os
import shutil
from pathlib import Path
from typing import Dict, List

from ktdk.asserts.matchers import IsNotEmpty, IsNotNone
from ktdk.core.mixins import FileUtilsMixin, ToolTaskMixin
from ktdk.core.tasks import Task

log = logging.getLogger(__name__)


def copytree(src, dst, symlinks=False, ignore=None):
    for item in os.listdir(src):
        source = os.path.join(src, item)
        destination = os.path.join(dst, item)
        if os.path.isdir(source):
            shutil.copytree(source, destination, symlinks, ignore)
        else:
            shutil.copy2(source, destination)


class FileActionWrapper:
    def __init__(self, action, path):
        self.action = action
        self.path = path

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return self.__dict__


class AbstractFilesTask(Task, FileUtilsMixin, ToolTaskMixin):
    @classmethod
    def _base_class(cls):
        return ToolTaskMixin

    def __init__(self, pattern='', subdir='', from_dir=None, to_dir=None, output_subdir=None,
                 source='submission', destination='workspace', overwrite=True, relative=True,
                 **kwargs):
        super().__init__(**kwargs)
        self.source = source
        self.destination = destination
        self.subdir = subdir or ''
        self.output_subdir = output_subdir or ''
        self._from_dir = Path(from_dir) if from_dir is not None else None
        self._to_dir = Path(to_dir) if to_dir is not None else None
        self.pattern = [pattern] if isinstance(pattern, str) else pattern
        self.overwrite = overwrite
        self.relative = relative

    @property
    def from_path(self) -> Path:
        if self._from_dir is not None:
            return self._from_dir
        return self.context.paths.get_dir(self.source) / self.subdir

    @property
    def to_path(self):
        if self._to_dir is not None:
            return self._from_dir
        return self.context.paths.get_dir(self.destination) / self.subdir / self.output_subdir

    def _glob_dirs(self, path=None) -> List:
        path = path or self.from_path
        found = []
        if self.pattern:
            for pat in self.pattern:
                found += path.glob(pat)
            return found
        return [path]

    @property
    def bound_params(self) -> Dict:
        return dict(pattern=self.pattern, source=self.source, target=self.destination,
                    subdir=self.subdir, subdir_source=self._from_dir, relative=self.relative,
                    subdir_target=self._to_dir, overwrite=self.overwrite)

    @property
    def bind_glob(self) -> Dict:
        return dict(pattern=self.pattern, source=self.source, subdir=self.subdir)


class CopyFiles(AbstractFilesTask):
    TOOL_NAME = 'copy'

    def _run(self, *args, **kwargs):
        log.info(f"[FS] Copy files {self.namespace}: {self.pattern}")
        self._file_utils.process_files(**self.bound_params, action=shutil.copy2)


class CopyDir(AbstractFilesTask):
    TOOL_NAME = 'copy_dir'

    def _run(self, *args, **kwargs):
        log.info(f"[FS] Copy dir {self.namespace}: {self.pattern}")
        self._file_utils.process_files(**self.bound_params, action=shutil.copytree)


class MoveFiles(AbstractFilesTask):
    TOOL_NAME = 'move'

    def _run(self, *args, **kwargs):
        log.info(f"[FS] Move files {self.namespace}: {self.pattern}")
        self._file_utils.process_files(**self.bound_params, action=shutil.move)


class DeleteFiles(AbstractFilesTask):
    TOOL_NAME = 'delete'

    def _run(self, *args, **kwargs):
        log.info(f"[FS] Delete files {self.namespace}: {self.pattern}")
        files = self._file_utils.glob_files(**self.bind_glob)
        for file_to_delete in files:
            log.debug(f"[DEL] Deleting file: {file_to_delete}")
            file_to_delete.unlink()


class ExistFiles(AbstractFilesTask):
    TOOL_NAME = ['exist', 'exists']

    def _run(self, *args, **kwargs):
        files = self._file_utils.glob_files(**self.bind_glob)
        log.info(f"[FS] Exists files {self.namespace}: {self.pattern}")
        self.asserts.check(
            files,
            IsNotEmpty(message=f'Files not found! Pattern {self.pattern} in {self.from_path}')
        )


class FindFiles(AbstractFilesTask):
    TOOL_NAME = 'find'

    def _run(self, *args, **kwargs):
        files = self._file_utils.glob_files(**self.bind_glob)
        log.debug(f"[FND] Found files with pattern={self.pattern}: {files}")
        self.context.config.add_task('found_files', files)
        self.asserts.check(files, IsNotNone(message=f"Cannot find any files: {self.pattern}"))


class MakeDir(AbstractFilesTask):
    TOOL_NAME = 'mkdir'

    def _run(self, *args, **kwargs):
        log.debug(f"[DIR] Create: {self.from_path}")
        self.from_path.mkdir(parents=True, exist_ok=True)



HELLO_DEF = """
config:
    timeout: 20

suite:
    name: hello
    description: Hello world intro program
    tasks:
        - fs:
              tool: copy
              submission:
                  - 'hello.c'
              test_files:
                  - pattern: 'solution.c'
        - build:
            tool: c_raw
            executable: student_hello
            files:
                - hello.c
            required: true
        - build:
            tool: c_raw
            executable: teacher_hello
            files:
                - solution.c
            required: true
    tests:
        - name: student_hello
          description: Student's hello should write hello
          tags:
            - naostro
          tasks:
            - execute:
                executable: student_hello
                executor: valgrind
                checks:
                    - type: stdout
                      regex: 'Hello World!'
                    - type: stderr
                      equals: ''
                    - type: return_code
                      equals: 0
"""



"""
Logging configuration module
"""
import logging
import os
import tempfile
from logging.config import dictConfig
from pathlib import Path


class Logging:
    def __init__(self, log_level_global=None, log_level_file=None,
                 log_level_console=None, results_dir=None, **kwargs):
        self._log_level_global = log_level_global or os.getenv('KTDK_LOG_LEVEL', 'INFO')
        self._log_level_file = log_level_file or os.getenv('KTDK_LOG_LEVEL_FILE',
                                                           self._log_level_global)
        self._log_level_console = log_level_console or os.getenv('KTDK_LOG_LEVEL_CONSOLE',
                                                                 self._log_level_global)
        self._results_dir = results_dir

    @property
    def log_level_global(self) -> str:
        return self._log_level_global

    @property
    def log_level_file(self) -> str:
        return self._log_level_file

    @property
    def log_level_console(self) -> str:
        return self._log_level_console

    @property
    def handlers(self):
        return {
            'console': self.get_handler_console(),
            'results_file': self.get_logger_file('results'),
        }

    @property
    def formatters(self):
        return {
            'verbose': {'format': '%(asctime)s - %(name)s - %(levelname)s - %(message)s'},
            'simple': {'format': '%(levelname)s %(message)s'},
            'colored_console': {
                '()': 'coloredlogs.ColoredFormatter',
                'format': "%(asctime)s - %(name)s - %(levelname)s - %(message)s",
                'datefmt': '%H:%M:%S'
            },
        }

    @property
    def loggers(self) -> dict:
        return {
            'ktdk': {
                'handlers': ['console', 'results_file'],
                'level': self.log_level_global, 'propagate': True
            },
            'tests': {'handlers': ['console'], 'level': 'DEBUG', 'propagate': True},
        }

    @property
    def logger_config(self):
        return {
            'version': 1,
            'disable_existing_loggers': False,
            'formatters': self.formatters,
            'handlers': self.handlers,
            'loggers': self.loggers,
        }

    @property
    def results_dir(self) -> Path:
        results = self._results_dir or tempfile.gettempdir()
        return Path(results)

    def get_logger_file(self, name, level: str = None):
        level = level or self.log_level_file
        results_dir = self.results_dir
        if not results_dir.exists():
            results_dir.mkdir(parents=True)
        return {
            'level': level,
            'class': 'logging.handlers.RotatingFileHandler',
            'formatter': 'verbose',
            'filename': str(self.results_dir / f'{name}.log'),
            'maxBytes': 5000000,  # 5MB
            'backupCount': 5
        }

    def get_handler_console(self, level: str = None):
        level = level or self.log_level_console
        return {
            'level': level, 'class': 'logging.StreamHandler', 'formatter': 'colored_console'
        }

    def load_config(self):
        """Loads config based on the config type
        Args:
        """
        add_custom_log_level()
        dictConfig(self.logger_config)
        reenable_loggers()


TRACE_LOG_LVL = 9


def _trace(self, message, *args, **kws):
    if self.isEnabledFor(TRACE_LOG_LVL):
        # Yes, logger takes its '*args' as 'args'.
        self._log(TRACE_LOG_LVL, message, args, **kws)


def add_custom_log_level():
    logging.addLevelName(TRACE_LOG_LVL, 'TRACE')
    logging.Logger.trace = _trace


def load_config(**config):
    return Logging(**config).load_config()


def reenable_loggers():
    loggers = [logging.getLogger(name) for name in logging.root.manager.loggerDict]
    for logger in loggers:
        logger.disabled = False



"""
Main KTDK module
"""
__version__ = '0.4.1'

from ktdk import log_config, utils

log_config.load_config()

import json
import logging
from pathlib import Path
from typing import Callable, List, Optional, Dict

import requests

from ktdk.config import Config
from ktdk.core import Result, Task, Test
from ktdk.runtime import Context
from ktdk.runtime.stat import stat_test

log = logging.getLogger(__name__)


def dump_to_file(context: Context, file_name: str, what):
    """Dumps content to the file
    Args:
        context(Context): Context instance
        file_name(str): File name
        what: What should be written to the file
    """
    safe_suffix = context.config.get('safe_suffixes')
    file_name = utils.unique_name(file_name, safe_suffixed=safe_suffix) + ".json"
    context.paths.save_result(file_name, json.dumps(what), raw=False)


class KTDK:
    """
    Main class of the KTDK, it holds configuration, the suite (root of the )
    """
    instance = None

    @staticmethod
    def get_instance(**cfg) -> 'KTDK':
        """Gets an instance of the KTDK
        Args:
            **cfg: KTDK config

        Returns(KTDK): KTDK instance

        """
        if KTDK.instance is None:
            log.debug(f"[KTDK] Instance config: {cfg}")
            KTDK.instance = KTDK(**cfg)
        return KTDK.instance

    def __init__(self, **cfg):
        """Creates instance of the KTDK suite
        Args:
            **cfg:
        """
        self.suite = Test(name='suite', tags=['suite'])
        self._post_actions: List[Callable] = []
        self.config = Config(**cfg)
        self.config.load()
        self.config['submission_config'] = self._parse_submission_config()
        log_config.load_config(**self.config, results_dir=self.config.paths.results)
        self.config.log_config()
        log.info(f"[KTDK] Version {__version__}")

    def register_tags(self, *tags):
        """Register avail. tags for the suite run
        Args:
            *tags: List of tags
        """
        if 'registered_tags' not in self.config:
            self.config['registered_tags'] = []
        self.config['registered_tags'].extend(tags)

    @property
    def post_actions(self) -> List[Callable]:
        """Gets all the post actions
        Returns(List[Callable]): List of all post actions
        """
        return self._post_actions

    def add_post_action(self, post_action: Callable):
        """Adds post action - action that will be executed after suite has been executed
        Args:
            post_action(Callable): Function that takes KTDK instance as the 1st arg and
                                    the Context as the 2nd arg

        """
        self._post_actions.append(post_action)

    def post_action(self):
        """Decorator that defines the post action
        Returns(Callable): Decorated Callable
        """

        def __post_action(func):
            self.add_post_action(func)

        return __post_action

    def create_context(self) -> Context:
        """Creates context from the configuration
        Returns(Context): The context instance
        """
        context = Context(suite_config=self.config, test_config={})

        return context

    def invoke(self) -> Result:
        """Invokes the KTDK suite
        Returns(Result): Overall result of the suite
        """
        return self.run()

    def run(self) -> Result:
        """Runs whole KTDK suite adn saves result

        Returns(Result): Overall result of the suite
        """
        context = self.create_context()
        log.info(f"[CTX] Created: {context}")
        suite_runner = self.suite.runner.get_instance(context=context)
        suite_result = suite_runner.invoke()
        self.__dump_required_files(context)
        self.__invoke_post_actions(context)
        return suite_result

    @property
    def stats(self) -> dict:
        return stat_test(self.suite)

    def __invoke_post_actions(self, context: Context):
        for post_action in self.post_actions:
            post_action(self, context)

        self.__send_notification(context=context)

    def __dump_required_files(self, context):
        if context.config['dump_result']:
            dump_to_file(context, 'suite-result', self.suite.to_dict())
        dump_to_file(context, 'suite-stats', self.stats)
        dump_to_file(context, 'suite-files', context.config['work_files'])

    def __send_notification(self, context: Context):
        webhook_url = context.config['webhook_url']
        if webhook_url:
            params = dict(url=webhook_url, data=json.dumps(self.stats))
            if context.config['webhook_token']:
                params['headers'] = {'Authorization': 'Bearer ' + context.config['webhook_token']}
            requests.post(**params)

    def _parse_submission_config(self) -> Optional[dict]:
        full_path = self._get_submission_config_path()
        if not full_path.exists():
            log.debug(f"[CONFIG] Submission config files has not been provided - {full_path}")
            return None
        sub_config = utils.parse_yaml_config(full_path)
        log.info(f"[LOAD] Loaded submission config {full_path}: {sub_config}")
        return sub_config

    def _get_submission_config_path(self) -> Optional[Path]:
        submission_config_file = self.config.get('submission_config_file')
        if submission_config_file is None:
            return None
        submission_path = self.config.get('submission')
        if submission_path is None:
            return None
        return Path(submission_path) / submission_config_file




from typing import Dict

import yaml

from ktdk import KTDK
from .base import DeclarativeSuiteLoader, DeclarativeTaskLoader, DeclarativeTestLoader
from .resolvers import BaseResolver, BuildTaskResolver, ExecuteTaskResolver, FSTasksResolver, \
    TaskBaseResolver, TestBaseResolver


def load_file(path) -> Dict:
    return yaml.safe_load(path)


def load_suite(definition, ktdk: KTDK) -> KTDK:
    return DeclarativeSuiteLoader(definition, ktdk).load()



import importlib
import importlib.machinery
import importlib.util
import logging
import os
import shutil
import sys
from pathlib import Path

from ktdk import Config, KTDK, utils, declarative
from ktdk.core import errors
from ktdk.declarative import registry

log = logging.getLogger(__name__)


def _import_definitions(path):
    parent_dir = path.parent
    if not path.exists():
        raise errors.FileNotExists(f"Definition file not exists: {path}")
    old = os.getcwd()
    os.chdir(parent_dir)
    schema = declarative.load_file(path)
    os.chdir(old)
    return declarative.load_suite(schema)


class CliManager:
    def __init__(self, config=None, **kwargs):
        self._params = dict(**kwargs)
        self._ktdk = None
        self._config_file = config

    @property
    def ktdk(self) -> KTDK:
        if self._ktdk is None:
            self._ktdk = self.get_ktdk()
        return self._ktdk

    @property
    def config(self) -> Config:
        return self.ktdk.config

    @property
    def workspace_dir(self) -> Path:
        return Path(self.ktdk.config['workspace']).absolute()

    @property
    def test_files_dir(self) -> Path:
        return Path(self.ktdk.config['test_files']).absolute()

    @property
    def submission_dir(self) -> Path:
        return Path(self.ktdk.config['submission']).absolute()

    @property
    def results_dir(self) -> Path:
        return Path(self.ktdk.config['results']).absolute()

    def load_project_files(self):
        """Loads the project files, adds the kontr_tests directly to the
        python modules, so the relative imports are working
        """
        module_path = self.test_files_dir / self.config.get('kontr_tests')
        sys.path.insert(0, str(module_path))
        path = module_path / self.config.get('entry_point')
        return _import_module_files('test_files', path)

    def load_test_definitions_files(self):
        """Loads the project files, adds the kontr_tests directly to the
        python modules, so the relative imports are working
        """
        module_path = self.test_files_dir / self.config.get('test_definitions')
        sys.path.insert(0, str(module_path))
        path = module_path / self.config.get('def_entry_point', 'suite.yml')
        return _import_definitions(path)

    def get_ktdk(self, **kwargs) -> KTDK:
        """Gets new instance of the KTDK
        Args:
            **kwargs: Parameters

        Returns(KTDK): Instance of the KTDK
        """
        config = {**self._params, **kwargs}
        log.debug(f"[CFG] Config: {config}")
        ktdk = KTDK.get_instance(**config)
        return ktdk

    def load_suite(self) -> KTDK:
        """Loads the suite -- it loads the scenario in the test files
        Returns(KTDK):

        """
        self.load_project_files()
        if self._config_file:
            self.ktdk.config.load_yaml(self._config_file)
        return self.ktdk

    def run_the_suite(self) -> KTDK:
        """Run the whole suite, call the invoke method on the KTDK instance
        Returns(KTDK): Instance of the KTDK
        """
        ktdk = self.load_suite()
        if self.ktdk.config.devel and self.ktdk.config.clean:
            workspace = self.ktdk.config.paths.workspace
            log.info(f"[CLEAN] Cleaning workspace: {workspace}")
            shutil.rmtree(workspace)
            utils.create_dirs(workspace)

        ktdk.invoke()
        return ktdk

    def print_registry(self, **kwargs):
        items = registry.full()
        for (reg_key, reg_val) in items.items():
            print(f"  {reg_key}")
            for (key, val) in reg_val.items():
                print(f"    {key}: {val}")


def _import_module_files(module_name: str, path: Path):
    """Imports and loads the scenario `instructions.py`F module in the kontr_tests dir
    Args:
        module_name(str): Name of the module,
        path:(Path): Location of the module
    Returns: Scenario definition module instance

    """
    full_path = str(path)
    loader = importlib.machinery.SourceFileLoader(module_name, full_path)
    spec = importlib.util.spec_from_loader(loader.name, loader)
    mod = importlib.util.module_from_spec(spec)
    loader.exec_module(mod)
    log.info(f"[CLI] Loading module: {mod}")
    return mod



from ktdk.tasks.command import Command

from ktdk.asserts import matchers, checks
from ktdk.scenarios import Scenario
from ktdk.tasks.cpp import ValgrindCommand
from ktdk.tasks.raw.executable import ExecutableTask


def _make_check(stream):
    return matchers.DiffFile(stream) if stream else matchers.IsEmpty()


class InOutFileScenario(Scenario):
    def __init__(self, executable=None, stdin=None, stdout=None, args=None, stderr=None,
                 executor=None, status_code=0, wd_subdir=None, **kwargs):
        self._stdin = stdin
        self._stdout = stdout
        self._stderr = stderr
        self._code = status_code
        self._args = args or []
        self._executor = executor or Command
        self._executable = executable
        self._wd_subdir = wd_subdir or ''
        super().__init__(**kwargs)

    def _run(self, *args, **kwargs):
        cfg = dict(input=self._stdin)
        cwd = self.context.paths.workspace / self._wd_subdir
        executable = ExecutableTask(self._executable, *self._args, cwd=cwd,
                                    executor=self._executor, command_config=cfg)
        stdout_check = _make_check(self._stdout)
        stderr_check = _make_check(self._stderr)
        executable.check_that(checks.StdOutMatchesCheck(stdout_check))
        executable.check_that(checks.StdErrMatchesCheck(stderr_check))
        executable.check_that(checks.ReturnCodeMatchesCheck(matchers.Equals(self._code)))
        self.root_test.add_task(executable)

        if self._executor == ValgrindCommand:
            executable.check_that(checks.ValgrindPassedCheck())



